############################################################################
#
# Copyright (c) 2007, 2008 Nexedi SARL and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

class ClientDisconnected(Exception):
  pass

class ExchangeProtocol:
  """
    Handle data exchange between client and server.
    Kinds of data which can be exchanged:
     - str
       send_field
       recv_field
     - int
       send_int
       recv_int
     - list of str
       send_list
       recv_list
     - list of int
       send_int_list
       recv_int_list
     - dict (key: str, value: int)
       send_dict
       recv_dict
    Forbidden chars:
      Send (raise if present):
        \\n (field separator)
      Receive (stripped silently):
        \\n (field separator)
        \\r (for compatibility)
  """
  def __init__(self, socket):
    self._socket = socket

  def send_field(self, to_send):
    if type(to_send) is not str:
      raise ValueError, 'Value is not of str type: %r' % (type(to_send), )
    if '\n' in to_send:
      raise ValueError, '\\n is a forbidden value.'
    self._socket.send(to_send)
    self._socket.send('\n')

  def recv_field(self):
    received = None
    result = []
    append = result.append
    while received != '\n':
      received = self._socket.recv(1)
      if len(received) == 0:
        raise ClientDisconnected
      if received != '\r':
        append(received)
    return ''.join(result[:-1])

  def send_int(self, to_send):
    self.send_field(str(to_send))

  def recv_int(self):
    return int(self.recv_field())

  def send_list(self, to_send, send_length=True):
    assert isinstance(to_send, (tuple, list))
    if send_length:
      self.send_int(len(to_send))
    for field in to_send:
      self.send_field(field)

  def send_int_list(self, to_send, *args, **kw):
    self.send_list([str(x) for x in to_send], *args, **kw)

  def recv_list(self, length=None):
    result = []
    append = result.append
    if length is None:
      length = int(self.recv_field())
    for field_number in xrange(length):
      append(self.recv_field())
    return result

  def recv_int_list(self, *args, **kw):
    return [int(x) for x in self.recv_list(*args, **kw)]

  def send_dict(self, to_send):
    """
      Key: string
      Value: int
    """
    assert isinstance(to_send, (dict))
    if len(to_send) == 0:
      key_list = value_list = []
    else:
      key_list, value_list = zip(*to_send.items())
    self.send_list(key_list)
    self.send_int_list(value_list, send_length=False)

  def recv_dict(self):
    """
      Key: string
      Value: int
    """
    key_list = self.recv_list()
    value_list = self.recv_int_list(len(key_list))
    result = dict(zip(key_list, value_list))
    return result




############################################################################
#
# Copyright (c) 2007, 2008 Nexedi SARL and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from ExchangeProtocol import ExchangeProtocol
from transaction._transaction import Transaction
from zLOG import LOG, WARNING, INFO, ERROR
import socket
import thread
import struct
import sys
import json

GET_LAST_COMMITED_TID_METHOD_ID = 'getLastCommitedTID'
from App.config import getConfiguration
config = getConfiguration()
# Note: TIDStorage is using simple <product-config tidstorage>
# as patching is done during importing product, so it is impossible
# to require %import Products.TIDStorage in zope.conf, as this code
# will be fired during importing
TID_STORAGE_ADDRESS = ('127.0.0.1', 9001)
if getattr(config, 'product_config', None) is not None:
  if not 'tidstorage' in config.product_config:
    LOG('TIDStorage', ERROR, 'Missing section <product-config tidstorage> in '
        'zope.conf')
  else:
    tidstorage_config = config.product_config['tidstorage']
    if not 'backend-ip' in tidstorage_config or not 'backend-port' in \
        tidstorage_config:
      LOG('TIDStorage', ERROR, 'Please configure backend-ip and backend-port'
          ' in <product-config tidstorage> section')
    else:
      try:
        TID_STORAGE_ADDRESS = (tidstorage_config['backend-ip'],
          int(tidstorage_config['backend-port']))
      except ValueError:
        LOG('TIDStorage', ERROR, 'backend-port shall be integer value, %r is '
            'wrong' % tidstorage_config['backend-port'])
else:
  LOG('TIDStorage', WARNING, 'product_config is not supported, hardcoded '
      'backend used.')

tid_storage = None
zope_identifier = None

LOG('TIDStorage',INFO,'Monkey patching transaction._transaction.Transaction.'
    '_commitResources, connected to %s' % ':'.join([str(q) for q in
      TID_STORAGE_ADDRESS]))

# Borrowed from CMFActivity.ActivityTool.getCurrentNode
def getZopeId():
  """ Return current node in form ip:port """
  global zope_identifier
  if zope_identifier is None:
    port = ''
    from asyncore import socket_map
    for k, v in socket_map.items():
      if hasattr(v, 'port') and hasattr(v, 'ip'):
        # see Zope/lib/python/App/ApplicationManager.py: def getServers(self)
        type = str(getattr(v, '__class__', 'unknown'))
        if type == 'ZServer.HTTPServer.zhttp_server':
          port = v.port
          ip = v.ip
          break
    assert port != '', 'zhttp_server not started yet'
    if TID_STORAGE_ADDRESS[0] != '127.0.0.1':
      assert ip != '127.0.0.1', 'self address must not be 127.0.0.1 if TIDStorage is remote'
    zope_identifier = '%s:%s' %(ip, port)
  return zope_identifier

def getFilestorageList(resource_list):
  return getFilestorageToTIDMapping(resource_list).keys()

def getFilestorageToTIDMapping(resource_list):
  datafs_tid_update_dict = {}
  for resource in resource_list:
    storage = getattr(resource, '_storage', None)
    if storage is not None:
      getLastCommitedTID = getattr(storage, GET_LAST_COMMITED_TID_METHOD_ID,
                                   None)
      if getLastCommitedTID is not None:
        tid = getLastCommitedTID()
        _addr = tuple([tuple(x) for x in getattr(storage, '_addr', [])])
        _storage = getattr(storage, '_storage', '')
        datafs_id = json.dumps((_addr, _storage))
        assert datafs_id not in datafs_tid_update_dict
        if tid is None:
          datafs_tid_update_dict[datafs_id] = None
        else:
          # unpack stolen from ZODB/utils.py:u64
          datafs_tid_update_dict[datafs_id] = struct.unpack(">Q", tid)[0]
  return datafs_tid_update_dict

class BufferedSocket:
  """
    Write-only thread-safe buffered socket.
    Attemps to reconnect at most once per flush.
  """

  _socket_lock = thread.allocate_lock()
  _connected = False
  
  def __init__(self, address):
    self._socket = socket.socket()
    self._address = address
    self._send_buffer_dict = {}

  def _connect(self):
    try:
      self._socket.connect(self._address)
      self._notifyConnected()
    except socket.error, message:
      # We don't want to have an error line per failed connection attemp, to
      # avoid flooding the logfile.
      pass

  def _getSendBuffer(self, ident):
    send_buffer = self._send_buffer_dict.get(ident)
    if send_buffer is None:
      send_buffer = self._send_buffer_dict[ident] = []
    return send_buffer

  def _notifyDisconnected(self, message):
    if self._connected:
      self._connected = False
      LOG('TIDStorage', WARNING, 'Disconnected: %s' % (message, ))

  def _notifyConnected(self):
    if not self._connected:
      self._connected = True
      # Display a log message at WARNING level, so that reconnection message
      # are visible when disconnection messages are visible, even if it is
      # not a warning, properly speaking.
      LOG('TIDStorage', WARNING, 'Connected')

  def send(self, to_send):
    send_buffer = self._getSendBuffer(thread.get_ident())
    send_buffer.append(to_send)

  def flush(self):
    """
      Flush send buffer and actually send data, with extra checks to behave
      nicely if connection is broken.
      Do not retry to send if something goes wrong (data is then lost !).
      Here, most important thing is speed, not data.
      Serialize usage.
    """
    ident = thread.get_ident()
    self._socket_lock.acquire()
    try:
      if not self._connected:
        self._connect()
      if self._connected:
        try:
          self._socket.sendall(''.join(self._getSendBuffer(ident)))
        except socket.error, message:
          self._notifyDisconnected(message)
          try:
            self._socket.shutdown(socket.SHUT_RDWR)
          except socket.error:
            self._socket.close()
          self._socket = socket.socket()
    finally:
      self._socket_lock.release()
    self._send_buffer_dict[ident] = []

class TIDClient:
  """Simple by design write only TIDClient using BufferedSocket"""
  def __init__(self, address):
    self._buffered_socket = BufferedSocket(address)
    self._field_exchange = ExchangeProtocol(socket=self._buffered_socket)

  def commit(self, tid_update_dict):
    """
      Send given dict to TIDStorage server.
    """
    self._send_command('commit')
    self._field_exchange.send_dict(tid_update_dict)
    self._buffered_socket.flush()

  def begin(self, storage_id_list):
    """
      Inform TIDStorage connection tracking that commit was initiated.
    """
    self._send_command('begin')
    self._field_exchange.send_list(storage_id_list)
    self._buffered_socket.flush()
  
  def abort(self):
    """
      Inform TIDStorage connection tracking that commit was aborted.
    """
    self._send_command('abort')
    self._buffered_socket.flush()

  def _send_command(self, command):
    """
      Every command must be followed by an identifier.
      This identifier is used to track transactions, so the same identifier
      must not be used twice at the same time, but can be reused later.
    """
    self._field_exchange.send_field(command)
    self._field_exchange.send_field('%s_%x' % (getZopeId(), thread.get_ident()))

original__commitResources = Transaction._commitResources
def _commitResources(self, *args, **kw):
  """
    Hook Transaction's _commitResources.

    Before:
     - Initialise TIDClient if needed
     - Check if there is any storage we are interested in in current commit
     - If so, issue a begin
    
    After (2 cases):
     - original__commitResources raised:
       - Issue an abort
     - otherwise:
       - Issue a commit

    Note to editors: Prevent your code from raising anything ! This method
    MUST NOT raise any exception, except that it MUST NOT hide any exception
    raised by original__commitResources.
    """
  has_storages = False
  try:
    global tid_storage
    if tid_storage is None:
      tid_storage = TIDClient(TID_STORAGE_ADDRESS)
    filestorage_list = getFilestorageList(self._resources)
    if len(filestorage_list):
      has_storages = True
      tid_storage.begin(filestorage_list)
  except:
    LOG('TIDStorage _commitResources', WARNING, 'Exception in begin phase', error=sys.exc_info())
  try:
    result = original__commitResources(self, *args, **kw)
  except:
    if has_storages:
      exception = sys.exc_info()
      try:
        tid_storage.abort()
      except:
        LOG('TIDStorage _commitResources', WARNING, 'Exception in abort phase', error=sys.exc_info())
      # Re-raise original exception, in case sendTIDCommitAbort tainted
      # last exception value.
      raise exception[0], exception[1], exception[2]
    else:
      raise
  else:
    if has_storages:
      # Now that everything has been commited, all exceptions relative to added
      # code must be swalowed (but still reported) to avoid confusing transaction
      # system.
      try:
        tid_storage.commit(getFilestorageToTIDMapping(self._resources))
      except:
        LOG('TIDStorage _commitResources', WARNING, 'Exception in commit phase', error=sys.exc_info())
  return result
 
Transaction._commitResources = _commitResources




#!/usr/bin/python2.4

##############################################################################
#
# Copyright (c) 2007 Nexedi SARL. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
# Parts of this file are borrowed from Zope 2.8.8 repozo.py script.
# Essentialy "usage" and "parseargs" methods.
# So it's released under the ZPL v2.0, as is Zope 2.8.8 .

""" repozo wrapper to backup for multiple Data.fs files and restore them in a consistent way.

Usage: %(program)s [-h|--help] [-c|--config configuration_file]
       [--repozo repozo_command] [-R|--recover|--recover_check]
       [--tid_log tid_log_file]
       [...]

  -h
  --help
    Display this help and exit.

  -c configuration_file
  --config configuration_file
    Use given file as configuration file.
    It must be a python file. See sample_configuration.py for required values.
    Recquired if neither -h nor --help are given.

  --repozo repozo_command
    Use given executable as repozo command.
    Default: repozo.py

  -R
  --recover
    Instead of saving existing Data.fs, perform an automated recovery from
    backups + timestamp file with optionally cutting files at the last coherent
    transaction.

  --recover_check
    Similar to above, except that it restores file to temp folder and compares
    with existing file.
    Files restored this way are automaticaly deleted after check.

  --recover_inplace
    Like 'recover' above, but doesn't actually restore from repozo files,
    instead it just cuts existing FileStorage files at the last coherent
    transaction.
  
  -t tid_log_file
  --tid_log tid_log_file
    TID log file, which will be used to find TID, which then will be used to
    cut restored file at found TID
"""

import imp
import getopt
import glob
import sys
import os
import md5
import time
import tempfile
from shutil import copy

from restore_tidstorage import parse, get_tid_position

program = sys.argv[0]

def log(message):
  print message

def cleanup(known_tid_storage_identifier_dict, keep_full_backup_count,
            status_file_backup_dir, status_file):
  if keep_full_backup_count <= 0:
    raise ValueError("A number of full backups to keep must be specified.")
  cleanup_list = []
  date_len = 19
  oldest_fsz = None # Oldest full backup to keep among all storages.
  for file_path, storage_path, object_path \
      in known_tid_storage_identifier_dict.itervalues():
    # Find oldest full backup to keep for this storage -> fsz
    fsz_list = sorted(glob.glob(os.path.join(storage_path, '*.fsz')))
    fsz = fsz_list[max(0, len(fsz_list) - keep_full_backup_count)]
    fsz, ext = os.path.splitext(os.path.basename(fsz))
    assert len(fsz) == date_len
    if oldest_fsz is None or fsz < oldest_fsz:
      oldest_fsz = fsz
    # Clean up all repozo files older than fsz.
    for path in glob.glob(os.path.join(storage_path, '*')):
      date, ext = os.path.splitext(os.path.basename(path))
      assert len(date) == date_len
      if ext in ('.fsz', '.deltafsz', '.dat') and date < fsz:
        cleanup_list.append(path)
  # Clean up all status files older than oldest_fsz.
  if oldest_fsz and status_file_backup_dir and status_file:
    prefix = os.path.join(status_file_backup_dir,
                          os.path.basename(status_file) + '-')
    path_len = len(prefix) + date_len
    for path in glob.glob(prefix + '*'):
      assert path.startswith(path) and len(path) == path_len
      if path[-date_len:] < oldest_fsz:
        cleanup_list.append(path)
  return cleanup_list

def backup(known_tid_storage_identifier_dict, repozo_formated_command):
  """Backups all ZODB files"""
  backup_count = 0
  total_count = len(known_tid_storage_identifier_dict)
  for key, (file_path, storage_path, object_path) in known_tid_storage_identifier_dict.iteritems():
    repozo_command = repozo_formated_command % (storage_path, file_path)
    if not os.access(storage_path, os.R_OK):
      os.makedirs(storage_path)
    log('Runing %r...' % (repozo_command, ))
    status = os.system(repozo_command)
    status = os.WEXITSTATUS(status)
    if status == 0:
      backup_count += 1
    else:
      log('Error occurred while saving %s: exit status=%i' % (file_path, status))
  log('Saved %i FileStorages out of %i.' % (backup_count, total_count))
  return total_count - backup_count

def get_md5_diggest(file_instance, length):
  BLOCK_SIZE=512
  file_instance.seek(0)
  md5sum = md5.new()
  read = file_instance.read
  update = md5sum.update
  while length > 0:
    to_read = min(BLOCK_SIZE, length)
    buffer = read(to_read)
    if len(buffer) != to_read:
      log('Warning: read %i instead of requested %i, stopping read' % (len(buffer), to_read))
      length = 0
    else:
      length -= to_read
    update(buffer)
  return md5sum.hexdigest()

def recover(known_tid_storage_identifier_dict, repozo_formated_command, check=False, last_tid_dict=None):
  """Recovers all ZODB files, when last_tid_dict is passed cut them at proper
     byte. If repozo_formated_command is None, assume we only need to trim the
     files to the last registered coherent transaction."""
  if repozo_formated_command is None and last_tid_dict is None:
    log('Error: we need the Status File for trimming the ZODB files')
    return 1
  recovered_count = 0
  total_count = len(known_tid_storage_identifier_dict)
  result = 0
  for key, (file_path, storage_path, object_path) in known_tid_storage_identifier_dict.iteritems():
    # check that indexes do not exists if cut shall be done
    file_index = file_path + ".index"
    if os.access(file_index, os.F_OK) and last_tid_dict is not None:
      log('Error: Index file %s exists. Bug in ZODB makes it impossible to cut'
          ' indexed file after restore. Cannot continue.' %
        file_index)
      result = 1
  if result:
    return result
  for key, (file_path, storage_path, object_path) in known_tid_storage_identifier_dict.iteritems():
    if repozo_formated_command is not None and not os.access(storage_path, os.R_OK):
      log('Warning: unable to recover %s because %s is missing/unreadable.' % (file_path, storage_path))
      continue
    elif repozo_formated_command is None and not (os.path.isfile(file_path) and
                                                  os.access(file_path, os.W_OK)):
      log('Warning: unable to trim %s because it is missing/unwritable' % (file_path,))
      continue
    if check:
      original_file_path = file_path
      file_path = os.path.join(tempfile.gettempdir(), os.path.basename(file_path))
    if repozo_formated_command is not None:
      repozo_command = repozo_formated_command % (storage_path, file_path)
      status = os.system(repozo_command)
      status = os.WEXITSTATUS(status)
    else:
      # we're being asked to cut only, so assume successful restore
      status = 0
    if status == 0:
      recovered_count += 1
      if last_tid_dict is not None:
        pos = get_tid_position(file_path, last_tid_dict[key])
        print 'Cutting restored file %s at %s byte' % (file_path, pos),
        f = open(file_path,'a')
        if not check:
          f.truncate(pos)
          print
        else:
          print 'only check, file untouched'
        f.close()
    else:
      log('Error occurred while recovering %s: exit status=%i' % (file_path, status))
    if check:
      log('Info: Comparing restored %s with original %s' % (file_path, original_file_path))
      recovered_file = open(file_path, 'r')
      original_file = open(original_file_path, 'r')
      try:
        recovered_file.seek(0, 2)
        original_file.seek(0, 2)
        recovered_file_length = recovered_file.tell()
        original_file_length = original_file.tell()
        checked_length = recovered_file_length
        if recovered_file_length < original_file_length:
          log('Info: Shorter than original: -%i bytes (-%.02f%%)' % \
              (original_file_length - recovered_file_length,
               1 - (float(recovered_file_length) / original_file_length)))
        elif recovered_file_length > original_file_length:
          log('ERROR: Longer than original: +%i bytes (+%.02f%%). Was original packed since backup ?' % \
              (recovered_file_length - original_file_length,
               float(recovered_file_length) / original_file_length))
          checked_length = None
        if checked_length is not None:
          recovered_file_diggest = get_md5_diggest(recovered_file, checked_length)
          original_file_diggest = get_md5_diggest(original_file, checked_length)
          if recovered_file_diggest != original_file_diggest:
            log('ERROR: Recovered md5 does not match original: %s != %s.' % \
                (recovered_file_diggest, original_file_diggest))
      finally:
        recovered_file.close()
        original_file.close()
      os.unlink(file_path)

  log('Restored %i FileStorages out of %i.' % (recovered_count, total_count))
  return total_count - recovered_count

def usage(code, msg=''):
  outfp = sys.stderr
  if code == 0:
    outfp = sys.stdout

  print >> outfp, __doc__ % globals()
  if msg:
    print >> outfp, msg

  sys.exit(code)

def parseargs():
  try:
    opts, args = getopt.getopt(sys.argv[1:], 'vQrt:FhzRc:',
                               ['help', 'verbose', 'quick', 'full',
                                'gzip', 'repository', 'repozo=',
                                'config=','recover', 'recover_check',
                                'recover_inplace', 'tid_log=', 'cleanup'])
  except getopt.error, msg:
    usage(1, msg)

  class Options:
    timestamp_file_path = None
    repozo_file_name = 'repozo.py'
    configuration_file_name = None
    repozo_opts = ['-B']
    known_tid_storage_identifier_dict = {}
    action = None
    dry_run = False
    inplace = False
    status_file = None
    status_file_backup_dir = None
    recover_status_file = None
    keep_full_backup_count = None

  options = Options()

  if args:
    options.repozo_opts.extend(args)

  for opt, arg in opts:
    if opt in ('-h', '--help'):
      usage(0)
    elif opt in ('-c', '--config'):
      options.configuration_file_name = arg
    elif opt == '--repozo':
      options.repozo_file_name = arg
    elif opt in ('-R', '--recover', '--recover_check', '--recover_inplace'):
      options.repozo_opts[0] = '-R'
      if options.action:
        usage(1, 'Only 1 command allowed.')
      options.action = 'recover'
      if opt == '--recover_check':
        options.dry_run = True
      if opt == '--recover_inplace':
        options.inplace = True
    elif opt in ('--cleanup'):
      if options.action:
        usage(1, 'Only 1 command allowed.')
      options.action = 'cleanup'
    elif opt in ('-r', '--repository'):
      options.repozo_opts.append('%s %s' % (opt, arg))
    elif opt in ('-t', '--tid_log'):
      options.recover_status_file = arg
    else:
      options.repozo_opts.append(opt)

  if options.configuration_file_name is None:
    usage(1, 'Either -c or --config is required.')

  configuration_filename, ext = os.path.splitext(os.path.basename(options.configuration_file_name))
  configuration_path = os.path.dirname(options.configuration_file_name)
  if len(configuration_path):
    configuration_path = [configuration_path]
  else:
    configuration_path = sys.path
  file, path, description = imp.find_module(configuration_filename, configuration_path)
  module = imp.load_module(configuration_filename, file, path, description)
  file.close()
  try:
    options.known_tid_storage_identifier_dict = module.known_tid_storage_identifier_dict
    options.timestamp_file_path = module.timestamp_file_path
  except AttributeError, msg:
    usage(1, msg)
  for option_id in ('status_file', 'status_file_backup_dir',
                    'keep_full_backup_count'):
    if getattr(options, option_id) is None:
      setattr(options, option_id, getattr(module, option_id, None))
  # XXX: we do not check any option this way, it's too dangerous.
  #options.repozo_opts.extend(getattr(module, 'repozo_opts', []))

  return options

def backupStatusFile(status_file,destination_directory):
  file_name = os.path.basename(status_file) + '-' + '%04d-%02d-%02d-%02d-%02d-%02d' % time.gmtime()[:6]
  copy(status_file, os.path.join(destination_directory, file_name))
  log("Written status file backup as %s" % os.path.join(destination_directory, file_name))
  
def main():
  options = parseargs()

  last_tid_dict = None
  if options.recover_status_file:
    if options.action != 'recover':
      raise ValueError("Status file path only for recovering")
    last_tid_dict = parse(options.recover_status_file)

  if options.action == 'cleanup':
    for path in cleanup(options.known_tid_storage_identifier_dict,
                        options.keep_full_backup_count,
                        options.status_file_backup_dir, options.status_file):
      os.remove(path)
    return 0

  repozo_formated_command = '%s %s -r "%%s"' % (options.repozo_file_name, ' '.join(options.repozo_opts))
  if options.action == 'recover':
    timestamp_file = open(options.timestamp_file_path, 'r')
    timestamp = ''
    read_line = ' '
    while len(read_line):
      timestamp = read_line
      read_line = timestamp_file.readline()
    timestamp = timestamp.strip('\r\n \t')
    if timestamp is not None:
      repozo_formated_command += ' -o "%%s" -D %s' % (timestamp, )
    if options.inplace:
      # use the Data.fs files themselves, instead of restoring them with repozo
      repozo_formated_command = None
    result = recover(
      known_tid_storage_identifier_dict=options.known_tid_storage_identifier_dict,
      repozo_formated_command=repozo_formated_command,
      check=options.dry_run,
      last_tid_dict=last_tid_dict)
  else:
    repozo_formated_command += ' -f "%s"'
    if options.status_file is not None and options.status_file_backup_dir is not None:
      backupStatusFile(options.status_file, options.status_file_backup_dir)
    result = backup(
      known_tid_storage_identifier_dict=options.known_tid_storage_identifier_dict,
      repozo_formated_command=repozo_formated_command)
    if result == 0:
      # Paranoid mode:
      # Issue a system-wide "sync" command to make sure all files which were saved
      # are really present on disk.
      os.system('sync')
      timestamp_file = open(options.timestamp_file_path, 'a', 0)
      try:
        # Borrowed from repozo.
        timestamp_file.write('\n%04d-%02d-%02d-%02d-%02d-%02d' % time.gmtime()[:6])
      finally:
        timestamp_file.close()
  return result

if __name__ == '__main__':
  sys.exit(main())



############################################################################
#
# Copyright (c) 2007 Nexedi SARL and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from ZEO.ClientStorage import ClientStorage
from zLOG import LOG, WARNING, INFO

LAST_COMMITED_TID_PROPERTY_ID = '_last_commited_tid'

# Hook tpc_finish's hook method.
# New hook must be a local method because it must access tpc_finish's "self"
# and original hook.

LOG('TIDStorage',INFO,'Monkey patching ClientStorage.tpc_finish and ClientStorage.getLastCommitedTID')

original_tpc_finish = ClientStorage.tpc_finish
def tpc_finish(self, txn, f=None):
  def saveTIDOnInstance(tid):
    if f is not None:
      f(tid)
    setattr(self, LAST_COMMITED_TID_PROPERTY_ID, tid)
  return original_tpc_finish(self, txn, f=saveTIDOnInstance)
ClientStorage.tpc_finish = tpc_finish

def getLastCommitedTID(self):
  """
    Return last commited tid for this storage, or None if no transaction
    was commited yet.
  """
  return getattr(self, LAST_COMMITED_TID_PROPERTY_ID, None)
ClientStorage.getLastCommitedTID = getLastCommitedTID




# COMMON
# This part is used both by tidstorage.py and repozo_tidstorage.py
known_tid_storage_identifier_dict = {
  "((('localhost', 8200),), '2')":
    ('/home/vincent/zeo2/var2/Data.fs',
     '/home/vincent/tmp/repozo/z22',
     'foo_test'),
  "((('localhost', 8200),), '1')":
    ('/home/vincent/zeo2/var/Data.fs',
     '/home/vincent/tmp/repozo/z21',
     'bar_test'),
  "((('localhost', 8100),), '1')":
    ('/home/vincent/zeo1/var/Data.fs',
     '/home/vincent/tmp/repozo/z11',
     'baz_test'),
}
base_url = 'http://login:password@localhost:5080/erp5/%s/modifyContext'
port = 9001
host = '127.0.0.1'

# SERVER
# This part is only used by server_v2.py
#logfile_name = 'tidstorage.log'
#pidfile_name = 'tidstorage.pid'
#fork = False
#setuid = None
#setgid = None
status_file = 'tidstorage.tid'
burst_period = 30
full_dump_period = 300

# REPOZO_TIDSTORAGE
# This part is only used by repozo_tidstorage.py
timestamp_file_path = 'repozo_tidstorage_timestamp.log'
# place to put backuped TIDStorage status_file logs
status_file_backup_dir = '/home/vincent/tmp/repozo'

# When cleaning up old backups, keep this number of full backups.
keep_full_backup_count = 3



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2009 Nexedi SARL and Contributors. All Rights Reserved.
#                    Vincent Pelletier <vincent@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

import socket
import time

from ExchangeProtocol import ExchangeProtocol

class TIDClient:
  def __init__(self, address):
    self._to_server = socket.socket()
    self._to_server.connect(address)
    self._exchange_protocol = ExchangeProtocol(self._to_server)

  def _dump(self, test_id=None):
    self._exchange_protocol.send_field('dump')
    received_dict = self._exchange_protocol.recv_dict()
    if test_id is None:
      result = received_dict
    else:
      id_len = len(test_id) + 1 # Add 1 to strip underscore.
      result = dict([(key[id_len:], value) \
                     for key, value in received_dict.iteritems() \
                     if key.startswith(test_id)])
    return dict([(key, int(value)) for key, value in result.iteritems()])

  def dump(self, test_id):
    return self._dump(test_id=test_id)

  def dump_all(self):
    return self._dump()

  def begin(self, test_id, transaction_id, storage_id_list):
    self._exchange_protocol.send_field('begin')
    self._exchange_protocol.send_field(transaction_id)
    internal_storage_id_list = ['%s_%s' % (test_id, x) \
                                for x in storage_id_list]
    self._exchange_protocol.send_list(internal_storage_id_list)

  def abort(self, test_id, transaction_id):
    self._exchange_protocol.send_field('abort')
    self._exchange_protocol.send_field(transaction_id)

  def commit(self, test_id, transaction_id, storage_tid_dict):
    self._exchange_protocol.send_field('commit')
    self._exchange_protocol.send_field(transaction_id)
    internal_storage_tid_dict = {}
    for key, value in storage_tid_dict.iteritems():
      internal_storage_tid_dict['%s_%s' % (test_id, key)] = value
    self._exchange_protocol.send_dict(internal_storage_tid_dict)

  def bootstraped(self):
    self._exchange_protocol.send_field('bootstraped')
    return self._exchange_protocol.recv_int()

  def waitForBootstrap(self):
    while not self.bootstraped():
      time.sleep(0.1)





#!/usr/bin/python2.4

##############################################################################
#
# Copyright (c) 2007 Nexedi SARL. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
# Parts of this file are borrowed from Zope 2.8.8 repozo.py script.
# Essentialy "usage", "parseargs" and parts of "restore" methods.
# So it's released under the ZPL v2.0, as is Zope 2.8.8 .

"""
Usage: %(program)s [-h|--help] [-c|--config configuration_file]

  -h
  --help
    Display this help and exit.

  -c configuration_file
  --config configuration_file
    Use given file as configuration file.
    It must be a python file.
    Required if neither -h nor --help are given.
"""

import imp
import getopt
import sys
import os
from struct import pack
import shutil
from ZODB.FileStorage import FileStorage

program = sys.argv[0]

def log(message):
  print message

def parse(status_file):
  tid_log = open(status_file)
  content = {}
  last_timestamp = None

  line = tid_log.readline()
  while line != '':
    split_line = line.split(' ', 2)
    assert len(split_line) == 3, repr(split_line)
    line_timestamp, line_type, line_dict = split_line
    line_timestamp = float(line_timestamp)
    assert line_type in ('f', 'd'), repr(line_type)
    if last_timestamp is None:
      last_timestamp = line_timestamp
    else:
      assert last_timestamp < line_timestamp, '%r < %r' % (last_timestamp, line_timestamp)
    line_dict = eval(line_dict, None)
    assert isinstance(line_dict, dict), type(line_dict)
    assert len(line_dict), repr(line_dict)
    if line_type == 'd':
      for key, value in line_dict.iteritems():
        if key in content:
          assert content[key] < value, '%r < %r' % (content[key], value)
        content[key] = value
    elif line_type == 'f':
      for key, value in content.iteritems():
        assert key in line_dict, repr(key)
        assert value <= line_dict[key], '%r <= %r' % (value, line_dict[key])
      content = line_dict
    line = tid_log.readline()
  return content

READCHUNK = 10 * 1024 * 1024

def get_tid_position(filepath,last_tid):
  tid = pack('>Q', last_tid + 1)
  # Find the file position of the last completed transaction.
  fs = FileStorage(filepath, read_only=True, stop=tid)
  # Note that the FileStorage ctor calls read_index() which scans the file
  # and returns "the position just after the last valid transaction record".
  # getSize() then returns this position, which is exactly what we want,
  # because we only want to copy stuff from the beginning of the file to the
  # last valid transaction record.
  pos = fs.getSize()
  fs.close()
  return pos

def recover(data_fs_backup_path_dict, status_file):
  last_tid_dict = parse(status_file)
  for storage_id, (file_path, backup_path) in data_fs_backup_path_dict.iteritems():
    # Derived from repozo (function=do_full_backup)
    # TODO: optimise to read backup only once.
    can_restore = False
    print "Considering:", storage_id
    if os.path.exists(backup_path):
      if os.path.exists(file_path):
        print 'Both original and backup files exist for %r. If previous restoration was successful, you should delete the backup for this restoration to take place. Original: %r Backup: %r' % (storage_id, file_path, backup_path)
      else:
        print 'Only backup file is available for %r: %r. Assuming it\'s ok and restoring to %r' % (storage_id, backup_path, file_path)
        can_restore = True
    else:
      if os.path.exists(file_path):
        sys.stdout.write('Copying %r to %r... ' % (file_path, backup_path))
        sys.stdout.flush()
        fsrc = file(file_path, 'rb')
        fdst = file(backup_path, 'wb')
        # shutil.copy uses too small a buffer size, which is slow:
        shutil.copyfileobj(fsrc, fdst, READCHUNK)
        fsrc.close()
        fdst.close()
        initial_size = os.stat(file_path).st_size
        final_size = os.stat(backup_path).st_size
        if initial_size == final_size:
          can_restore = True
          print 'Done.'
        else:
          print 'Backup size %i differs from original size %i. Is the original file (%r) still in use ? Is there enough free disk space at destination (%r) ?' % (final_size, initial_size, file_path, backup_path)
      else:
        print 'Cannot find any file for %r: %r and %r do not exist.' % (storage_id, file_path, backup_path)
    if can_restore:
      tid = last_tid_dict[storage_id]
      sys.stdout.write('Scanning %r for tid %r... ' % (backup_path, tid))
      sys.stdout.flush()
      pos = get_tid_position(backup_path,last_tid_dict[storage_id])
      print 'Done. Position:', pos
      original_file = open(file_path, 'rb+')
      original_file.truncate(pos)
      original_file.close()
      print 'Truncated %s to: %s bytes (transaction %r)' % (file_path, pos, tid)
    else:
      print 'Skipping restoration of %r (%r).' % (file_path, storage_id)

def usage(code, msg=''):
  outfp = sys.stderr
  if code == 0:
    outfp = sys.stdout

  print >> outfp, __doc__ % globals()
  if msg:
    print >> outfp, msg

  sys.exit(code)

def parseargs():
  try:
    opts, args = getopt.getopt(sys.argv[1:], 'hc:',
                               ['help', 'config='])
  except getopt.error, msg:
    usage(1, msg)

  class Options:
    configuration_file_name = None
    status_file = None

  options = Options()

  for opt, arg in opts:
    if opt in ('-h', '--help'):
      usage(0)
    elif opt in ('-c', '--config'):
      options.configuration_file_name = arg

  if options.configuration_file_name is None:
    usage(1, 'Either -c or --config is required.')

  configuration_filename, ext = os.path.splitext(os.path.basename(options.configuration_file_name))
  configuration_path = os.path.dirname(options.configuration_file_name)
  if len(configuration_path):
    configuration_path = [configuration_path]
  else:
    configuration_path = sys.path
  file, path, description = imp.find_module(configuration_filename, configuration_path)
  module = imp.load_module(configuration_filename, file, path, description)
  file.close()
  try:
    options.data_fs_backup_path_dict = module.data_fs_backup_path_dict
    options.status_file = module.status_file
  except AttributeError, msg:
    usage(1, msg)
  return options

def main():
  options = parseargs()
  recover(
    data_fs_backup_path_dict=options.data_fs_backup_path_dict,
    status_file=options.status_file)

if __name__ == '__main__':
  sys.exit(main())



#!/usr/bin/python

import traceback
import sys

from TIDClient import TIDClient

class TestTIDServerV2:
  def __init__(self, address, port):
    self._client = TIDClient((address, port))
 
  def assertEqual(self, value, target):
    assert value == target, 'value %r does not match target %r' % (value, target)
 
  def test_01_InitialValue(self, test_id):
    """
      Check that the storage is empty
    """
    self.assertEqual(self._client.dump_all(), {})
  
  def test_02_Bootstrap(self, test_id):
    """
      Trigger bootstrap and check that no value is visible until bootstrap is
      done.
    """
    t1_storage_tid_dict = {'s0': 1}
    t2_storage_tid_dict = {'s1': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    # Bootstrap is runing on the server, nothing is visible yet.
    self.assertEqual(self._client.dump(test_id), {})
    self._client.waitForBootstrap()
    # Nothing is available yet, we need one more transaction to happen.
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    # Now everything must be available.
    self.assertEqual(self._client.dump(test_id), {'s0': 1, 's1': 1})

  def test_03_Scenario1(self, test_id):
    """
      Simple begin - commit case.
    """
    storage_tid_dict = {'s1': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), storage_tid_dict)

  def test_04_Scenario2(self, test_id):
    """
      Simple begin - abort case.
    """
    storage_tid_dict = {'s1': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.abort(test_id, 't1')
    self.assertEqual(self._client.dump(test_id), {})

  def test_05_Scenario3(self, test_id):
    """
      2 concurent transactions impacting a common storage.
      Second transaction begins after first does, and commits before
      first does.
    """
    t1_storage_tid_dict = {'s1': 1, 's2': 1}
    t2_storage_tid_dict = {'s1': 2, 's3': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 2, 's2': 1, 's3': 1})

  def test_06_Scenario4(self, test_id):
    """
      3 concurent transactions.
      Transactions 1 and 2 impact same storage s1.
      Transaction 3 impacts storage s3 after transaction 2 commited.
      Still, as storage 3  was part of a non-commitable-yet transaction,
      it must not be commited untill all blockable (here, t1) transaction have
      ended.
    """
    t1_storage_tid_dict = {'s1': 1, 's2': 2}
    t2_storage_tid_dict = {'s1': 2, 's3': 1}
    t3_storage_tid_dict = {'s3': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't3', t3_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't3', t3_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 2, 's2': 2, 's3': 1})

  def test_07_Scenario4bis(self, test_id):
    """
      3 concurent transactions.
      Transactions 1 and 2 impact same storage s1.
      Transaction 3 impacts storage s3 after transaction 2 commited.
      Still, as storage 3  was part of a non-commitable-yet transaction,
      it must not be commited untill all blockable (here, t1) transaction have
      ended.
      In this version, t1 aborts: for example, tpc_vote failed. As the data
      was already sent to storage, it might be already present on disk (and
      anyway, tid is not to be used anymore), so it's valid for t2 to commit
      with tid 2 even if t1 aborted tid 1.
    """
    t1_storage_tid_dict = {'s1': 1, 's2': 2}
    t2_storage_tid_dict = {'s1': 2, 's3': 1}
    t3_storage_tid_dict = {'s3': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't3', t3_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't3', t3_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.abort(test_id, 't1')
    self.assertEqual(self._client.dump(test_id), {'s1': 2, 's3': 1})

  def test_08_Scenario5(self, test_id):
    """
      2 concurent transactions impacting a common storage.
      Second transaction begins after first does, and commits after
      first does.
    """
    t1_storage_tid_dict = {'s1': 2}
    t2_storage_tid_dict = {'s1': 1, 's2': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 2, 's2': 1})

  def test_09_Scenario6(self, test_id):
    """
      2 concurent transactions impacting separate sets of storages.
      Check that the first commit impacts dump data immediately.
    """
    t1_storage_tid_dict = {'s1': 1}
    t2_storage_tid_dict = {'s2': 1}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 1})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 1, 's2': 1})

  def test_10_Scenario7(self, test_id):
    """
      3 concurent transactions.
      t1 and t2 impact a set of different storages.
      t3 impacts a set of storage containing the ones from t1 and the ones
      from t2.
      Check that nothing impacts dump data until everything is commited.
    """
    t1_storage_tid_dict = {'s1': 1}
    t2_storage_tid_dict = {'s2': 2}
    t3_storage_tid_dict = {'s1': 2, 's2': 2}
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't2', t2_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.begin(test_id, 't3', t3_storage_tid_dict.keys())
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't1', t1_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't2', t2_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {})
    self._client.commit(test_id, 't3', t3_storage_tid_dict)
    self.assertEqual(self._client.dump(test_id), {'s1': 2, 's2': 2})

  def test_11_Scenario8(self, test_id):
    """
      Simple increase case.
    """
    self.assertEqual(self._client.dump(test_id), {})
    t1_storage_tid_dict = {}
    for s1_value in (1, 2):
      previous_t1_storage_tid_dict = t1_storage_tid_dict
      t1_storage_tid_dict = {'s1': s1_value}
      self._client.begin(test_id, 't1', t1_storage_tid_dict.keys())
      self.assertEqual(self._client.dump(test_id), previous_t1_storage_tid_dict)
      self._client.commit(test_id, 't1', t1_storage_tid_dict)
      self.assertEqual(self._client.dump(test_id), t1_storage_tid_dict)

  def run(self):
    for test_method_id in [x for x in dir(self) if x.startswith('test')]:
      self.log("Runing %s..." % (test_method_id, ))
      try:
        try:
          getattr(self, test_method_id)(test_id=test_method_id)
        except AssertionError:
          self.log('F\n')
          self.log('\n'.join(traceback.format_exception(*sys.exc_info())))
      finally:
        self.log('\n')

  def log(self, message):
    sys.stdout.write(message)

if __name__ == '__main__':

  assert len(sys.argv) == 3, 'Requires exactly 2 arguments: <address> <port>'

  address = sys.argv[1]
  port = int(sys.argv[2])

  test = TestTIDServerV2(address, port)
  test.run()




# See http://peak.telecommunity.com/DevCenter/setuptools#namespace-packages
try:
        __import__('pkg_resources').declare_namespace(__name__)
except ImportError:
        from pkgutil import extend_path
        __path__ = extend_path(__path__, __name__)



from setuptools import setup, find_packages
import os

product_path = os.path.join('Products', 'TIDStorage')

version = open(os.path.join(product_path, 'VERSION.txt')).read().split()[-1]

setup(name='Products.TIDStorage',
      version=version,
      description="TIDStorage Product provides a way to have consistent "\
           "backups when running a multi-storage instance (only ZEO is "\
           "supported at the moment).",
      long_description=open(os.path.join(product_path, 'README')).read() + "\n" + \
                       open(os.path.join(product_path, 'CHANGES')).read(),
      classifiers=[
        "Framework :: Zope2",
        "Operating System :: OS Independent",
        ],
      keywords='Zope TIDStorage',
      author='Vincent Pelletier',
      author_email='vincent@nexedi.com',
      url='https://svn.erp5.org/repos/public/erp5/trunk/utils/Products.TIDStorage/',
      packages=find_packages(),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'Zope2',
      ],
      entry_points={
        'console_scripts': [
          'tidstoraged = Products.TIDStorage.bin.tidstorage:main',
          'tidstorage_repozo = Products.TIDStorage.repozo.repozo_tidstorage:main',
          'tidstorage_restore = Products.TIDStorage.repozo.restore_tidstorage:main',
        ]
      }
      )
